require 'faraday'
require 'json/ext'
require 'map'
module Tatami
  class Couch
    def conn
      @conn ||= Faraday.new(:url => @url) do |faraday|
        faraday.request  :url_encoded             # form-encode POST params
        faraday.response :logger                  if @opts[:log]
        faraday.adapter  Faraday.default_adapter  # make requests with Net::HTTP
      end
    end
    # delete and recreate the database
    def blank!
      conn.delete "/#{db_env}"
      conn.put "/#{db_env}"
      self
    end

    attr_reader :db
    def db_env()
      @env ? "#{@db}_#{@env}" : @db
    end

    # @option opts [Boolean] :log
    # @option opts [String] :db
    def initialize db, url='http://localhost:5984', opts={}
      @env = opts.delete(:env)
      @db, @url, @opts = db, url, opts
    end

    def get path, opts={}
      puts "GET #{path}" if opts[:log]
      data = conn.get(path).body
      cast_response data, opts[:format]
    end

    def cast_response data, format
      case format
      when :string
        data
      when :object
        JSON.parse(data)
      else
        Map.new JSON.parse(data)
      end
    end

    def get_db_string path
      conn.get("/#{db_env}#{path}").body
    end

    def get_db path, opts={}
      get "/#{db_env}#{path}", opts
    end

    def verb method, url, body, opts={}
      puts "#{method.upcase} #{url} #{body.inspect}" if opts[:log]
      resp = conn.send method, url, JSON.generate(body), 'Content-Type' => 'application/json'
      cast_response resp.body, opts[:format]
    end

    def post_db url, body, opts={}
      verb :post, "/#{db_env}#{url}", body, opts
    end

    def put_db(url, body)  verb :put,  "/#{db_env}#{url}", body end

    def post_db_string url, body
      conn.post("/#{db_env}#{url}", JSON.generate(body), 'Content-Type' => 'application/json').body
    end

    def version() get('/')[:version] end
    def db_info() get_db "" end
#db_name: cars
#doc_count: 0
#doc_del_count: 0
#update_seq: 0
#purge_seq: 0
#compact_running: false
#disk_size: 79
#data_size: 0
#instance_start_time: '1364126614255010'
#disk_format_version: 6
#committed_update_seq: 0

    def with_fixtures! names, *docs
      blank!
      load_design
      documents = docs.map do |values|
        Hash[ names.zip(values) ]
      end
      doc_new_bulk documents
    end

    def with_documents! *docs
      blank!
      load_design
      doc_new_bulk docs
    end

    # PUT /db/doc
    # @return [Doc]
    # ok  true
    # id  0c8e32db863b87c9f6233d572854d6db
    # rev 1-08c111cb2ccfbf117ed931bef444f730
    def doc_new id, attributes
      id ||= next_id
      put_db "/#{id}", attributes
    end

    # @param [String] file_root
    def doc_attach_app id, file_root
      # curl -X PUT -v http://127.0.0.1:5984/cars/flot/index.html --data-binary "@test/fixtures/index.html" --header "Accept: text/html" --header "Transfer-Encoding: chunked" --header "Expect:"
      require 'tatami/app'
      put_db "/#{id}", "_attachments" => App.new.attachments(file_root)
    end

    def doc_create_or_update id, attributes
      if old = doc(id)
        doc_update old, attributes
      else
        doc_new id, attributes
      end
    end

    # @return [String] autogenerated available id (caches 10)
    def next_id
      @next_ids ||= []
      if @next_ids.empty?
        @next_ids = get("_uuids?count=10")[:uuids]
      end
      @next_ids.pop
    end

    # _id  0c8e32db863b87c9f6233d5728545922
    # _rev 1-08c111cb2ccfbf117ed931bef444f730
    # error: not_found
    # reason: missing
    # name Frank
    # GET /db/doc
    def doc id
      r = get_db "/#{id}"
      r[:error] == "not_found" ? nil : r
    end

    # GET /db/_design/design-doc/_show/show-name
    # GET /db/_design/design-doc/_show/show-name/doc
    def show name, doc=nil, design=db_env
      d = doc ? "/#{doc}" : ""
      get_db_string "/_design/#{design}/_show/#{name}#{d}"
    end

    # @param [Hash] old
    # @param [Hash] new
    def doc_update old, new
      attributes = new.merge(:_rev => old[:_rev])
      doc_new old[:_id], attributes
    end

    def doc_merge_cached id, attributes
      @cache ||= Map.new
      @cache[id] ||= begin
        doc(id) || Map.new
      end
      @cache.combine! id => attributes
      r = doc_new id, @cache[id]
      @cache[id][:_rev] = r[:rev]
    end

    # creates new type in _design/[db-name] document
    # @param [Symbol] type :view, :list, :show
    def design_new type, name, content
      doc_merge_cached "_design/#{db_env}", :language => 'coffeescript',
                                         "#{type}s" =>
                                            { name => content }
    end

# views name map
# views name reduce
# lists name
    def load_design pattern="*"
      [ %w[erlang %% erl],
        %w[coffeescript # coffee]
      ].each do |language, comment, ext |
        Dir.glob("couches/#{@db}/#{pattern}.#{ext}").each do |path|
          source = IO.read(path)
          name = File.basename(path)[/[^\.]+/]
          map = Map.new(language: language)
          source.split(/^(?=#{comment} (?:views|lists) \S+)/m).each do |element|
            element.split("\n").first =~ /#{comment} (\S+) (\S+) ?(map|reduce)?/
            values = [$1,$2,$3].compact << element
            map.set *values
          end
          doc_create_or_update "_design/#{name}", map
        end
      end
    end

    def doc_merge id, attributes
      old = doc id
      doc_new old[:_id], old.combine(attributes)
    end

    def doc_new_bulk documents
      post_db "/_bulk_docs", {:docs => documents}
    end

    def bulk_import limit=100, &b
      BulkImport.new self, limit, &b
    end

    class BulkImport
      def initialize couch, limit
        @couch, @limit, @cache = couch, limit, []
        yield self
        docs_new
      end

      def doc_new attributes
        docs_new if @cache.size > @limit
        @cache << attributes
      end

      def docs_new
        @couch.doc_new_bulk @cache
        @cache = []
      end
    end

    # GET /db/_design/design-doc/_view/view-name
    def view name
      @views ||= Map.new
      @views[name] ||= View.new self, name.to_s
    end

    class View
      def initialize couch, name
        @couch = couch
        name.prepend "#{@couch.db}/" unless name['/']
        @design, @view = name.split('/')
        @base = "/_design/#{@design}/_view/#{@view}"
      end

      # @param [] value
      # @return [String] json
      def options_to_param opts
        if single_key=opts.delete(:singlekey)
          opts[:startkey] =  single_key
          opts[:endkey]   = (single_key + [{}] )
        end
        opts.map{|option,value| "#{option}=#{ JSON.generate([value])[1..-2] }"}.join('&')
      end

      # {total_rows: 1
      #  offset: 0
      #  rows: [{id: 45d73462c193b4bd5b5e643b33010de0
      #          key: the_key
      #          value: the_value}, ...
      #        ]

      # @option opts [Boolean] :format  :string, :object, (:map)
      # @option opts [Boolean] :log  puts url
      # @option opts [Array] :keys make a post request
      # @option opts [Array] :single_key calculates startkey and endkey
      # @return [Map]
      def data opts={}
        opts_add = {log: opts.delete(:log), format: opts.delete(:format) }
        list = opts[:list] ? "_list/#{opts.delete(:list)}" : "_view"
        if opts[:keys]
          @couch.post_db "/_design/#{@design}/#{list}/#{@view}", opts, opts_add
        else
          @couch.get_db "/_design/#{@design}/#{list}/#{@view}?#{options_to_param opts}", opts_add
        end
      end

      def rows opts={}
        data(opts)[:rows]
      end

      # @param [String] map
      # @p
      # @return [View]
      def create map, reduce=nil
        v = {:map => map}
        v[:reduce] = reduce if reduce
        @couch.design_new :view, @view, v
        self
      end

      def key_values opts={}
        rows(opts).map{|r| [ r[:key],r[:value] ]}
      end

      def values opts={}
        rows(opts).map{|d| d[:value]}
      end

      def keys opts={}
        rows(opts).map{|d| d[:key]}
      end

      # @return value of the first row
      def reduced opts={}
        data(opts).get :rows, 0, :value
      end

      # @return [Map] rows as key => value
      def reduce opts={}
        rows(opts).each_with_object(Map.new)do |row,o|
          o[row[:key]] = row[:value]
        end
      end

      # special function for arrays as keys
      def reduced_value *values
        reduced :group_level => values.size, :startkey => values, :endkey => values.dup << {}
      end

    end
  end
end
